package com.example.ueberwachungssystem.VideoDetection;

import android.content.Context;
import android.graphics.ImageFormat;
import android.media.Image;
import android.util.Log;
import android.util.Size;

import androidx.annotation.NonNull;
import androidx.camera.core.CameraSelector;
import androidx.camera.core.ExperimentalGetImage;
import androidx.camera.core.ImageAnalysis;
import androidx.camera.core.ImageProxy;
import androidx.camera.core.Preview;
import androidx.camera.lifecycle.ProcessCameraProvider;
import androidx.camera.view.PreviewView;
import androidx.core.content.ContextCompat;
import androidx.lifecycle.LifecycleOwner;

import com.google.common.util.concurrent.ListenableFuture;

import java.nio.ByteBuffer;
import java.util.concurrent.ExecutionException;


@ExperimentalGetImage
public class VideoDetector {

    // Calling Activity
    private final Context context;
    // Camera Provider
    private ProcessCameraProvider cameraProvider;
    private Boolean isDetectionRunning = false;
    // Preview Camera Image
    private PreviewView previewView = null;
    // Check Violation
    private final float DELTA_LUMINOSITY_THRESHOLD = 0.3f;
    private Float previousLuminosity = null;
    // On Detection Listener
    private OnDetectionListener listener;



    /** Constructor */
    public VideoDetector(Context context) {
        this.context = context;
    }


    /** On Detection Listener - runs when a violation is reported */
    public interface OnDetectionListener {
        void onDetection(DetectionReport detectionReport);
    }
    public void setOnDetectionListener(OnDetectionListener listener) {
        this.listener = listener;
    }
    public void reportViolation(float amplitude) {
        if (listener != null) {
            DetectionReport detectionReport = new DetectionReport("123", "Video", amplitude);
            listener.onDetection(detectionReport);
        }
    }


    /** Return State of Video Detector */
    public Boolean isRunning() {
        return isDetectionRunning;
    }


    /** Starts Video Detection */
    public void startDetection() {
        if (isDetectionRunning)
            return;
        // Request Camera Provider
        final ListenableFuture<ProcessCameraProvider> cameraProviderFuture = ProcessCameraProvider.getInstance(context);
        //Check for Camera availability
        cameraProviderFuture.addListener(new Runnable() {
            @Override
            public void run() {
                try {
                    cameraProvider = cameraProviderFuture.get();
                    bindLuminosityAnalysis(cameraProvider);
                    isDetectionRunning = true;
                } catch (ExecutionException | InterruptedException e) {
                    // No errors need to be handled for this Future. This should never be reached.
                }
            }
        },ContextCompat.getMainExecutor(context));
    }


    /** Stops Video Detection */
    public void stopDetection() {
        if (!isDetectionRunning)
            return;
        cameraProvider.unbindAll();
        isDetectionRunning = false;
    }


    /** Set PreviewView to show Image */
    public void setPreviewView(PreviewView previewView) {
        this.previewView = previewView;
    }


    /** Binds the Luminosity Analyzer (configure and run Analysis) */
    private void bindLuminosityAnalysis(@NonNull ProcessCameraProvider cameraProvider) {
        // Configure and create Image Analysis
        ImageAnalysis.Builder builder = new ImageAnalysis.Builder();
        builder.setTargetResolution(new Size(640, 480));
        builder.setBackpressureStrategy(ImageAnalysis.STRATEGY_KEEP_ONLY_LATEST);
        ImageAnalysis imageAnalysis = builder.build();

        // Set Analyzer
        imageAnalysis.setAnalyzer(ContextCompat.getMainExecutor(context), new ImageAnalysis.Analyzer() {
            @Override
            public void analyze(@NonNull ImageProxy imageProxy) {
                if (imageProxy.getFormat() == ImageFormat.YUV_420_888) {
                    Image image = imageProxy.getImage();
                    assert image != null;

                    // Analyze frame
                    float luminosity = calculateLuminosity(image);
                    Log.d("Video Detector", String.valueOf(luminosity));
                    if (previousLuminosity != null)
                        checkForViolation(luminosity, previousLuminosity);
                    previousLuminosity = luminosity;
                }
                imageProxy.close();
            }
        });
        // Create Preview
        Preview preview = new Preview.Builder().build();
        // Specify which Camera to use
        CameraSelector cameraSelector = new CameraSelector.Builder().requireLensFacing(CameraSelector.LENS_FACING_BACK).build();
        // Update PreviewView if set
        if (previewView != null)
            preview.setSurfaceProvider(previewView.getSurfaceProvider());

        cameraProvider.bindToLifecycle((LifecycleOwner) context, cameraSelector, imageAnalysis, preview);
        stopDetection();
    }


    /** Return Luminosity from Image */
    private float calculateLuminosity (Image image) {
        int width = image.getWidth();
        int height = image.getHeight();

        Image.Plane[] planes = image.getPlanes();
        ByteBuffer luminosityBuffer = planes[0].getBuffer();

        int yRowStride = image.getPlanes()[0].getRowStride();
        int yPixelStride = image.getPlanes()[0].getPixelStride();

        int luminosity;
        float sum = 0;

        for (int y = 0; y < height; ++y) {
            for (int x = 0; x < width; x++) {
                int index = (y * yRowStride) + (x * yPixelStride);
                luminosity = (luminosityBuffer.get(index) & 0xff);
                sum += luminosity;
            }
        }
        return sum / (width * height);
    }


    /** Check if delta Luminosity exceeds threshold */
    private void checkForViolation(float luminosity, float previousLuminosity) {
        float deltaLuminosity = Math.abs(luminosity - previousLuminosity);
        if (deltaLuminosity > DELTA_LUMINOSITY_THRESHOLD) {
            reportViolation(deltaLuminosity);
            Log.d("Violation", "Violation");
        }
        Log.d("Delta", String.valueOf(Math.abs(previousLuminosity - luminosity)));
    }
}
